/*
** iplog_util.c - iplog utility functions.
** Copyright (C) 1999 Odin <odin@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
*/

#include <config.h>

#define _GNU_SOURCE
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <syslog.h>
#include <netdb.h>
#include <ctype.h>
#include <stdarg.h>
#include <string.h>
#include <pwd.h>
#include <grp.h>
#include <iplog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>

#include <iplog_options.h>
#include <iplog_dns.h>

#if !defined(HAVE_GETHOSTBYADDR_R) || !defined(HAVE_GETPROTOBYNUMBER_R) || \
	!defined(HAVE_GETSERVBYPORT_R)
#	include <pthread.h>
#endif

extern u_long flags;

#ifndef HAVE_GETHOSTBYADDR_R
pthread_mutex_t host_mutex = PTHREAD_MUTEX_INITIALIZER; 
#endif

#ifndef HAVE_GETSERVBYPORT_R
pthread_mutex_t serv_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif

#ifndef HAVE_GETPROTOBYNUMBER_R
pthread_mutex_t proto_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif

#ifndef HAVE_DPRINTF
int dprintf(int fd, const char *fmt, ...) {
	va_list ap;
	ssize_t ret;
#ifdef HAVE_VASPRINTF
	u_char *buf;

	va_start(ap, fmt);
	vasprintf((char **) &buf, fmt, ap);
#else /* ! HAVE_VASPRINTF */
	u_char buf[2048];

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
#endif /* ! HAVE_VASPRINTF */
	va_end(ap);
	ret = write(fd, buf, strlen(buf));
#ifdef HAVE_VASPRINTF
	free(buf);
#endif /* HAVE_VASPRINTF */

	return (ret);
}
#endif /* ! HAVE_DPRINTF */

void *xmalloc(size_t size) {
	void *ret = malloc(size);

	if (ret == NULL)
		fatal("Fatal: Out of memory.");

	return (ret);
}

void *xcalloc(size_t nmemb, size_t size) {
	void *p = calloc(nmemb, size);

	if (p == NULL)
		fatal("Fatal: Out of memory.");

	return (p);
}

char *xstrdup(const char *s) {
	char *ret = strdup(s);

	if (ret == NULL)
		fatal("Fatal: Out of memory.");

	return (ret);
}

char *xstrncpy(char *dest, const char *src, size_t n) {
	u_char *ret = dest;

	if (n == 0)
		return (dest);

	while (--n > 0 && (*dest++ = *src++) != '\0')
		;
	*dest = '\0';

	return (ret);
}

char *xstrncat(char *dest, const char *src, size_t n) {
	u_char *ret = dest;

	if (n == 0)
		return (dest);

	for (; *dest != '\0' ; dest++, n--)
		;
	while (--n > 0 && (*dest++ = *src++) != '\0')
		;
	*dest = '\0';

	return (ret);
}

u_char *serv_lookup(u_short port, const u_char *proto, u_char *buf, size_t len) {
#ifdef HAVE_GETSERVBYPORT_R
	struct servent se;
#ifndef REENTRANT_FUNCTIONS_IN_LIBPTHREAD
	struct servent *temp;
#endif
	u_char rbuf[1024];

#ifndef REENTRANT_FUNCTIONS_IN_LIBPTHREAD
	if (getservbyport_r(port, proto, &se, rbuf, sizeof(rbuf), &temp) != 0 ||
		temp == NULL)
#else
	if (getservbyport_r(port, proto, &se, rbuf, sizeof(rbuf)) == NULL)
#endif
		snprintf(buf, len, "port %u", htons(port));
	else
		xstrncpy(buf, se.s_name, len);
#else
	struct servent *se;

	pthread_mutex_lock(&serv_mutex);
	se = getservbyport(port, proto);
	if (se != NULL)
		xstrncpy(buf, se->s_name, len);
	else {
		snprintf(buf, len, "port %u", htons(port));
	}
	pthread_mutex_unlock(&serv_mutex);
#endif

	return (buf);
}

u_char *inet_ntoa_r(struct in_addr *in, u_char *buf, size_t len) {
	u_long addr = ntohl(in->s_addr);

	snprintf(buf, len, "%lu.%lu.%lu.%lu",
			((addr >> 24) & 0xff), ((addr >> 16) & 0xff),
			((addr >> 8) & 0xff), (addr & 0xff));

	return (buf);
}

u_char *proto_lookup(int proto, u_char *buf, size_t len) {
#ifndef HAVE_GETPROTOBYNUMBER_R
	struct protoent *pe;

	pthread_mutex_lock(&proto_mutex);
	pe = getprotobynumber(proto);
	if (pe != NULL)
		xstrncpy(buf, pe->p_name, len);
	else
		snprintf(buf, len, "proto %u", proto);
	pthread_mutex_unlock(&proto_mutex);
#else
	struct protoent pe;
#ifndef REENTRANT_FUNCTIONS_IN_LIBPTHREAD
	struct protoent *temp;
#endif
	u_char tbuf[256];

#ifndef REENTRANT_FUNCTIONS_IN_LIBPTHREAD
	if (getprotobynumber_r(proto, &pe, tbuf, sizeof(tbuf), &temp) != 0 ||
		temp == NULL)
#else
	if (getprotobynumber_r(proto, &pe, tbuf, sizeof(tbuf)) == NULL)
#endif
		snprintf(buf, len, "proto %u", proto);
	else
		xstrncpy(buf, pe.p_name, len);
#endif

	return (buf);
}

u_char *host_lookup(struct in_addr *in, bool resolv, u_char *buf, size_t len) {
	if (resolv == false)
		return (inet_ntoa_r(in, buf, len));

	if (flags & DNS_CACHE)
		return (get_dns_cache(in->s_addr, buf, len));

	return (_host_lookup(in, buf, len));
}

u_char *_host_lookup(struct in_addr *in, u_char *buf, size_t len) {
	u_char *ret = NULL;
#ifdef HAVE_GETHOSTBYADDR_R
	struct hostent he;
	u_char hbuf[1024];
	int herr;
#ifndef REENTRANT_FUNCTIONS_IN_LIBPTHREAD
	struct hostent *temp;
#endif /* ! REENTRANT_FUNCTIONS_IN_LIBPTHREAD */

	if (gethostbyaddr_r((char *) in, sizeof(struct in_addr), AF_INET,
#ifndef REENTRANT_FUNCTIONS_IN_LIBPTHREAD
		&he, hbuf, sizeof(hbuf), &temp, &herr) != 0 || temp == NULL
#else /* REENTRANT_FUNCTIONS_IN_LIBPTHREAD */
		&he, hbuf, sizeof(hbuf), &herr) == NULL
#endif /* ! REENTRANT_FUNCTIONS_IN_LIBPTHREAD */
		|| (strlen(he.h_name) >= len))
		ret = inet_ntoa_r(in, buf, len);
	else {
		xstrncpy(buf, he.h_name, len);
		ret = buf;
	}
#else /* ! HAVE_GETHOSTBYADDR_R */
	struct hostent *he;

	pthread_mutex_lock(&host_mutex);
	he = gethostbyaddr((char *) in, sizeof(struct in_addr), AF_INET);

	if (he != NULL && strlen(he->h_name) < len)
		ret = xstrncpy(buf, he->h_name, len);
	else
		ret = inet_ntoa_r(in, buf, len);

	pthread_mutex_unlock(&host_mutex);
#endif /* ! HAVE_GETHOSTBYADDR_R */

	return (ret != NULL ? ret : (u_char *) "UNKNOWN");
}

void fork_to_back(void) {
	int f = fork();

	if (f > 0)
		_exit(0);
	else if (f != 0)
		fatal("Fatal: fork(): %s", strerror(errno));

	setsid();
	chdir("/");
	umask(022);

	close(0);
	close(1);
	close(2);

	f = open("/dev/null", O_WRONLY);
	if (f != 0)
		fatal("Fatal: open: /dev/null: %s", strerror(errno));
	dup2(0, 1);
	dup2(0, 2);
}

void drop_privs(const u_char *user, const u_char *group) {
	uid_t uid;
	uid_t gid;

	if (group != NULL) {
		if (valid_number(group))
			gid = strtoul(group, NULL, 10);
		else {
			struct group *gr = getgrnam(group);
			if (gr == NULL)
				fatal("Fatal: Unknown group: \"%s\"", group);
			gid = gr->gr_gid;
		}

		if (setgid(gid) == -1)
			fatal("Fatal: setgid(%d): %s", gid, strerror(errno));

		if (setgroups(0, NULL) == -1)
			fatal("Fatal: setgroups(0, NULL): %s", strerror(errno));
	}

	if (user != NULL) {
		if (valid_number(user))
			uid = strtoul(user, NULL, 10);
		else {
			struct passwd *pw = getpwnam(user);
			if (pw == NULL)
				fatal("Fatal: Unknown user: \"%s\"", user);
			uid = pw->pw_uid;
		}

		if (setuid(uid) == -1)
			fatal("Fatal: setuid(%d): %s", uid, strerror(errno));
	}
}

bool valid_number(const u_char *p) {
	if (*p == '0')
		return (p[1] == '\0');
	if (*p == '-')
		p++;
	if (*p == '\0' || *p == '0')
		return (false);
	for (; *p != '\0' ; p++) {
		if (!isdigit(*p))
			return (false);
	}

	return (true);
}

void fatal(const u_char *fmt, ...) {
	va_list ap;
#ifdef HAVE_VASPRINTF
	u_char *message;

	va_start(ap, fmt);
	vasprintf((char **) &message, fmt, ap);
#else
	u_char message[2048];

	va_start(ap, fmt);
	vsnprintf(message, sizeof(message), fmt, ap);
#endif

	va_end(ap);

	if (isatty(STDERR_FILENO))
		fprintf(stderr, "%s\n", message);
	else
		mysyslog(LOG_WARNING, message);

	exit(-1);
}

int get_line(FILE *fp, u_char *buf, size_t len) {
	int c;
	size_t i = 0;

	for (;;) {
		c = getc(fp);

		switch (c) {
			case EOF:
				if (i == 0)
					return EOF;
			case '\n':
				if (i >= len)
					i = 0;
				buf[i] = '\0';
				return (0);
			default:
				if (i == 0 && isspace(c))
					break;
				if (i < len)
					buf[i++] = c;
		}
	}

	return (0);
}

void *list_append(const void *data, void **head, size_t len) {
	struct generic_data *cur, *temp = xmalloc(len);

	memcpy(temp, data, len);
	temp->next = NULL;

	if (*head == NULL) {
		*head = temp;
		return (*head);
	}

	cur = *head;
	while (cur->next != NULL)
		cur = cur->next;
	cur->next = temp;

	return (temp);
}

void *list_prepend(const void *data, void **head, size_t len) {
	struct generic_data *save, *temp;

	temp = xmalloc(len);
	memcpy(temp, data, len);

	save = *head;
	*head = temp;
	temp->next = save;

	return (temp);
}

void *list_delete(void *node, void **head) {
	struct generic_data *ret = NULL, *cur = *head, *temp = NULL;

	while (cur != NULL) {
		if (cur == node) {
			ret = cur->next;

			if (temp == NULL)
				*head = ret;
			else
				temp->next = ret;

			free(cur);
			break;
		}

		temp = cur;
		cur = cur->next;
	}

	return (ret);
}

void list_destroy(void *list_head, void (*cleanup)(void *)) {
	struct generic_data *cur = list_head, *next;

	if (cleanup == NULL) {
		while (cur != NULL) {
			next = cur->next;
			free(cur);
			cur = next;
		}
	} else {
		while (cur != NULL) {
			next = cur->next;
			cleanup(cur);
			free(cur);
			cur = next;
		}
	}
}

/* vim:ts=4:sw=8:tw=0 */
